//
//  main.c
//  SEM_movement
//
//  Created by Alexander Gord on 7/5/11.
//  Last Updated 6/11/2014
//  Copyright 2011-2014
//
// Usage: SEM_movement <initial conditions file> <options file> <potential list> <output file>
// example: SEM_movement single_cell_IC options potentials2 polar.001
// Automatically generated outputs: center_output.txt outputs cell center data for each time step
// cell_output.txt outputs cell number data for each time step

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/stat.h>
#include <math.h>
#include <string.h>
#include <time.h>


#include <OpenCL/OpenCL.h>
#include "const.h"

// reads the options file specified and loads the data into variables
void read_options(const char* filename, int* iterations, int* max_elements, char* output_format, int* iterations_per_frame)
{
    FILE *fp;
    
    fp=fopen(filename, "r");
    
    fscanf(fp, "%d %d %s %d",iterations, max_elements, output_format, iterations_per_frame);
    printf("Iterations: %d\nMax Elements: %d\n", *iterations, *max_elements);
    
    fclose(fp);
}

// reads the initial conditions file specified and loads the data into the element arrays
void read_initial_conditions(const char* filename, int* num_element_types, int* num_element_arrays, const int max_elements, float ***subcellx, float ***subcelly, float ***subcellz, int ***cell_num,int **num_elements,int **element_type) 
{
    // create a file pointer
    FILE *fp;
    int length = 255;
    char *line;
    int itr = 0; // an iterator
    
    
    line = (char *)malloc((length+1)*sizeof(char));
    
    
    fp = fopen(filename, "r");
    if(fp == NULL)
    {
        fprintf(stderr,"Can't open %s!\n",filename);
        exit(1);
    }
 
    fscanf(fp, "%d", num_element_arrays);
#ifdef CELL_GROWTH
    *num_element_arrays=*num_element_arrays+1;
#endif
    
#ifdef CPU_CELL_GROWTH
    *num_element_arrays=*num_element_arrays+1;
#endif
    
    printf("Using %d element arrays\n", *num_element_arrays);
    
    
    *subcellx = (float**)malloc(sizeof(float*) * (*num_element_arrays));
    *subcelly = (float**)malloc(sizeof(float*) * (*num_element_arrays));
    *subcellz = (float**)malloc(sizeof(float*) * (*num_element_arrays));
    *cell_num = (int**)malloc(sizeof(int*) * (*num_element_arrays));
    *num_elements = (int*)malloc(sizeof(int) * (*num_element_arrays));
    *element_type = (int*)malloc(sizeof(int) * (*num_element_arrays));
    
    for(int i = 0; i < *num_element_arrays; i++)
    {
        (*subcellx)[i]=(float *)malloc(max_elements*sizeof(float));
        (*subcelly)[i]=(float *)malloc(max_elements*sizeof(float));
        (*subcellz)[i]=(float *)malloc(max_elements*sizeof(float));
        (*cell_num)[i]=(int *)malloc(max_elements*sizeof(int));
    }
    
    while((fscanf(fp, "%d %d", &((*element_type)[itr]), ((*num_elements)+itr))) != EOF)
    {
        printf("Array %d has elements of type %d and contains %d elements\n",itr, (*element_type)[itr], (*num_elements)[itr]);
        
        printf("#    x   y   z\n");
        
        for(int i = 0; i<(*num_elements)[itr]; i++)
        {
            fscanf(fp, "%d %f %f %f", &((*cell_num)[itr][i]), &((*subcellx)[itr][i]),&((*subcelly)[itr][i]), &((*subcellz)[itr][i]));
            printf("%d %f %f %f\n", (*cell_num)[itr][i], (*subcellx)[itr][i], (*subcelly)[itr][i], (*subcellz)[itr][i]);
        }
        itr++;
    }

    
#ifdef CELL_GROWTH
    (*num_elements)[*num_element_arrays-1]=0;

    
    int temp_num_element_types = 0;
    for(int i = 0; i<(*num_element_arrays-1); i++)
    {
        if(temp_num_element_types<(*element_type)[i])
        {
            temp_num_element_types=(*element_type)[i];
        }
    }
    
    (*element_type)[*num_element_arrays-1]=temp_num_element_types+1;
    *num_element_types=temp_num_element_types+2;
#else

#ifdef CPU_CELL_GROWTH
    (*num_elements)[*num_element_arrays-1]=0;
    
    
    int temp_num_element_types = 0;
    for(int i = 0; i<(*num_element_arrays-1); i++)
    {
        if(temp_num_element_types<(*element_type)[i])
        {
            temp_num_element_types=(*element_type)[i];
        }
    }
    
    (*element_type)[*num_element_arrays-1]=temp_num_element_types+1;
    *num_element_types=temp_num_element_types+2;
#else
    
    
    int temp_num_element_types = 0;
    for(int i = 0; i < *num_element_arrays; i++)
    {
        if(temp_num_element_types<(*element_type)[i])
        {
            temp_num_element_types=(*element_type)[i];
        }
    }
    *num_element_types=temp_num_element_types+1;
#endif
#endif

        
    fclose(fp);
    /*
    for( int i=0; i<*num_element_arrays; i++)
    {
        for( int j=0; j<(*num_elements)[i]; j++)
        {
            printf("x=%2.2f, y=%2.2f, z=%2.2f, cell number=%i\n", (*subcellx)[i][j], (*subcelly)[i][j], (*subcellz)[i][j], (*cell_num)[i][j]);
        }
    }
    */
    
    
    return;

}

// reads the potentials file and loads the data into variables
void read_potential_functions(const char * file, int *num_params, float **params, int num_element_types, char*** filenames)
{
    // create a file pointer
    FILE *fp;
    int length = 255;
    char *line;
    
        
    
    line = (char *)malloc((length+1)*sizeof(char));
    
    
    fp = fopen(file, "r");

    
    if(fp == NULL)
    {
        fprintf(stderr,"Can't open %s!\n",file);
        exit(1);
    }

    
    
    fscanf(fp, "%d", num_params);
    printf("Using %d parameters\n", *num_params);

    
    
    
    *params=(float*)malloc((*num_params) * sizeof(float));
    
    
    for(int i = 0; i<*num_params; i++)
    {
        fscanf(fp,"%f",&((*params)[i]));
        printf("Parameter %d is set to %f\n", i, (*params)[i]);
    }
    
     
    int type_1, type_2;
    char* filename;
    filename = (char*)malloc(1024);

    printf("type1 type2 filename\n");
    
    while((fscanf(fp, "%d %d %s", &type_1, &type_2, filename)) != EOF)
    {
        if(type_1<num_element_types && type_2<num_element_types)
        {
            strcpy(filenames[type_1][type_2],filename);
            printf("%d %d %s\n",type_1,type_2,filenames[type_1][type_2]);
        }
    }


    
    
 
    
    fclose(fp);
    
    return;
   
}

// a function to write the positions of all elements into a file that can be read as an initial condition
void write_positions( FILE* fp, int num_element_arrays, float **subcellx, float **subcelly, float **subcellz, int **cell_num, int *num_elements,int *element_type) 
{
        if(fp == NULL)
    {
        fprintf(stderr,"Invalid output file!\n");
        exit(1);
    }
    
    fprintf(fp, "%d\n", num_element_arrays);
        
    for(int i = 0; i < num_element_arrays; i++)
    {
        fprintf(fp, "%d %d\n", element_type[i], num_elements[i]);
        for(int j = 0; j<num_elements[i]; j++)
        {
            fprintf(fp, "%d %f %f %f\n", cell_num[i][j], subcellx[i][j], subcelly[i][j], subcellz[i][j]);
        }
    }

    
    fclose(fp);
    
    return;
    
}

// writes the positions of the elements into a file that can be read by gnuplot
void write_positions_gnuplot( FILE* fp, int num_element_arrays, float **subcellx, float **subcelly, float **subcellz, int **cell_num, int *num_elements,int *element_type) 
{
    if(fp == NULL)
    {
        fprintf(stderr,"Invalid output file!\n");
        exit(1);
    }
    
    for(int i = 0; i < num_element_arrays; i++)
    {
        for(int j = 0; j<num_elements[i]; j++)
        {
            fprintf(fp, "%d %d %f %f %f\n",element_type[i], cell_num[i][j], subcellx[i][j], subcelly[i][j], subcellz[i][j]);
        }
    }
    
    fclose(fp);
    
    return;
    
}

// writes the poisitions of the elements into a pqr format that can be read by VMD
void write_positions_pqr( FILE* fp, int num_element_arrays, float **subcellx, float **subcelly, float **subcellz, int **cell_num, int *num_elements,int *element_type) 
{
    if(fp == NULL)
    {
        fprintf(stderr,"Invalid output file!\n");
        exit(1);
    }
    int entry_num=0;
    for(int i = 0; i < num_element_arrays; i++)
    {
        for(int j = 0; j<num_elements[i]; j++)
        {   
            entry_num++; //this is intended to start at 1
            fprintf(fp, "ATOM %6d C  THR   %5d      %6.3f %6.3f %6.3f %5.3f %5.3f\n", entry_num, cell_num[i][j], subcellx[i][j], subcelly[i][j], subcellz[i][j],0.1*(float)element_type[i],1.0);
        }
    }
    
    fclose(fp);
    
    return;
    
}


// loads a program source file so OpenCL can compile it on the fly
char * load_program_source(const char *filename)
{ 
	
	struct stat statbuf;
	FILE *fh; 
	char *source; 
    
    printf("Reading Source %s\n",filename);
	
	fh = fopen(filename, "r");
	if (fh == 0)
		return 0; 
	
	stat(filename, &statbuf);
	source = (char *) malloc(statbuf.st_size + 1);
	fread(source, statbuf.st_size, 1, fh);
	source[statbuf.st_size] = '\0'; 
	
	return source; 
} 

// calculate the distance between two elements
float distance(const float x1, const float y1, const float z1, const float x2, const float y2, const float z2)
{
    return sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)+(z1-z2)*(z1-z2));
}

// moves an element from one element type to another element type
void move_element( int i, int j, int new_element_type, int max_elements, int *num_elements, float **subcellx, float **subcelly, float **subcellz, int **cell_num, int new_cell_num)
{
    // i is the element type of the element to be moved, j is its position in the array
    if( num_elements[new_element_type]==max_elements)
    {
        printf("Maximum number of elements reached\n");
        return;
    }
    
    subcellx[new_element_type][num_elements[new_element_type]]=subcellx[i][j];
    subcelly[new_element_type][num_elements[new_element_type]]=subcelly[i][j];
    subcellz[new_element_type][num_elements[new_element_type]]=subcellz[i][j];
    cell_num[new_element_type][num_elements[new_element_type]]=new_cell_num;
    num_elements[new_element_type]++;
    
    for(int k=j; k<num_elements[i]-1; k++)
    {
        subcellx[i][k]=subcellx[i][k+1];
        subcelly[i][k]=subcelly[i][k+1];
        subcellz[i][k]=subcellz[i][k+1];
        cell_num[i][k]=cell_num[i][k+1];
    }
    num_elements[i]--;
    
    
}

// delaminates a cell moving all of its elements into the suprabasal element type
void cell_delaminate(int cell_id, int num_element_types, int num_element_arrays, int *element_type, int max_elements, int *num_elements, float **subcellx, float **subcelly, float **subcellz, int **cell_num, int new_cell_num)
{
    
    for( int i = 0; i<2; i++ )
    {
        for( int j = 0; j<num_elements[i]; j++ )
        {
            if(cell_num[i][j]==cell_id)
            {
                move_element(i,j,2,max_elements, num_elements, subcellx, subcelly, subcellz, cell_num, new_cell_num);
                j--;
            }
        }
    }
    return;
    
}

// deletes an element
void delete_element( int i, int j, int *num_elements, float **subcellx, float **subcelly, float **subcellz, int **cell_num)
{
    // i is the element type of the element to be deleted, j is its position in the array
    
    
    for(int k=j; k<num_elements[i]-1; k++)
    {
        subcellx[i][k]=subcellx[i][k+1];
        subcelly[i][k]=subcelly[i][k+1];
        subcellz[i][k]=subcellz[i][k+1];
        cell_num[i][k]=cell_num[i][k+1];
    }
    num_elements[i]--;
    
}


// deletes every element in the given cell
void delete_cell(int cell_id, int *num_elements, float **subcellx, float **subcelly, float **subcellz, int **cell_num)
{
    
    for( int i = 0; i<3; i++ )
    {
        for( int j = 0; j<num_elements[i]; j++ )
        {
            if(cell_num[i][j]==cell_id)
            {
                delete_element(i,j, num_elements, subcellx, subcelly, subcellz, cell_num);
                j--;
            }
        }
    }
    return;
    
}

// This is the funcion that does the bulk of the computing calculating the motion of the elements and executing cell growth and division
void exec_kernels(int num_element_types, int num_element_arrays, int *element_type, int max_elements, int *num_elements, float **subcellx, float **subcelly, float **subcellz, int **cell_num, float dt, int max_itr, int num_params, float *params, char*** filenames, char* output_name, int iterations_per_frame )
{
    
    float dt_2 = 2.0f * dt; // we will use this much later 
    
    //allocate memory to store contexts, devices, cammand queues and errors
    cl_context  context;
    cl_command_queue cmd_queue;
    cl_device_id device;
    
    cl_int err;
    
    //initiate variables that will be used to generate noise. The lower and upper bounds
    //set the amplitude of the noise
    unsigned int random_seed;
    time_t timer;
    float lower_bound= -0.005f;
    float upper_bound=  0.005f;
    

    //create arrays to store programs and kernels
    char *program_source, *potential_source, *full_program_source;
    char *noisy_copy_source, *cell_growth_source, *cell_division_source;
    char *program_name = DEFAULT_PROGRAM;
        
    cl_program **programs;
    cl_kernel **kernels;
    
    cl_program noisy_copy_program,cell_growth_program, cell_division_program;
    cl_kernel noisy_copy_kernel, cell_growth_kernel, cell_division_kernel;
    
    //Initialize random numbers
    srand(time(NULL));
    
    programs = (cl_program **)malloc(num_element_types*sizeof(cl_program*));
    kernels = (cl_kernel **)malloc(num_element_types*sizeof(cl_kernel*));
        
    program_source = load_program_source(program_name);
    
    for(int i = 0; i<num_element_types; i++)
    {
        programs[i] = (cl_program*)malloc(num_element_types*sizeof(cl_program));
        kernels[i] = (cl_kernel*)malloc(num_element_types*sizeof(cl_kernel));
    }
   
    
    err = clGetDeviceIDs(NULL, COMPUTE_DEVICE, 1, &device, NULL);
    
    size_t returned_size = 0;
    cl_char vendor_name[1024];
    cl_char device_name[1024];
    err = clGetDeviceInfo(device, CL_DEVICE_VENDOR, sizeof(vendor_name), vendor_name, &returned_size);
    err = clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(device_name), device_name, &returned_size);
    
    printf("Connecting to %s %s...\n", vendor_name, device_name);
    
    context = clCreateContext(0, 1, &device, NULL, NULL, &err);
    assert(err==CL_SUCCESS);
    cmd_queue = clCreateCommandQueue(context, device, 0, NULL);

    char build[2048];
    
    printf("Loading noisy copy program source: add_noise.cl\n");
    noisy_copy_source = load_program_source("add_noise.cl");
    noisy_copy_program = clCreateProgramWithSource(context, 1, (const char**)&noisy_copy_source, NULL, &err);
    assert(err == CL_SUCCESS);
    err = clBuildProgram(noisy_copy_program, 0, NULL, NULL, NULL, NULL);
    printf("%d\n",err);
    assert(err == CL_SUCCESS);
    printf("Build Log:\n%s\n",build);
    noisy_copy_kernel = clCreateKernel(noisy_copy_program, "add_noise", &err);
    assert(err == CL_SUCCESS);
    
    printf("Loading cell growth program: cell_growth.cl\n");
    cell_growth_source = load_program_source("cell_growth.cl");
    cell_growth_program = clCreateProgramWithSource(context, 1, (const char**)&cell_growth_source, NULL, &err);
    assert(err == CL_SUCCESS);
    err = clBuildProgram(cell_growth_program, 0, NULL, NULL, NULL, NULL);
    printf("%d\n",err);
    assert(err == CL_SUCCESS);
    printf("Build Log:\n%s\n",build);
    cell_growth_kernel = clCreateKernel(cell_growth_program, "cell_growth", &err);
    assert(err == CL_SUCCESS);

    printf("Loading cell division program: cell_division.cl\n");
    cell_division_source = load_program_source("cell_division.cl");
    cell_division_program = clCreateProgramWithSource(context, 1, (const char**)&cell_division_source, NULL, &err);
    assert(err == CL_SUCCESS);
    err = clBuildProgram(cell_division_program, 0, NULL, NULL, NULL, NULL);
    printf("%d\n",err);
    assert(err == CL_SUCCESS);
    printf("Build Log:\n%s\n",build);
    cell_division_kernel = clCreateKernel(cell_division_program, "cell_division", &err);
    assert(err == CL_SUCCESS);

    
    
    for(int i = 0; i<num_element_types; i++)
    {
        for(int j = 0; j<num_element_types; j++)
        {
            printf("Loading program for types %d and %d: '%s'\n", i, j, filenames[i][j]);
            potential_source = load_program_source(filenames[i][j]);
            full_program_source = (char*)malloc(strlen(potential_source)+strlen(program_source)+2);
            sprintf(full_program_source, "%s %s", potential_source, program_source);
            // printf("%s\n",full_program_source);
            programs[i][j]=clCreateProgramWithSource(context,1, (const char**)&full_program_source, NULL, &err);
            assert(err==CL_SUCCESS);
            
            err = clBuildProgram(programs[i][j], 0, NULL, NULL, NULL, NULL);
            assert(err==CL_SUCCESS);
            
            clGetProgramBuildInfo(programs[i][j], device, CL_PROGRAM_BUILD_LOG, 2048, build, NULL);
            printf("Build Log:\n%s\n",build);
            
            kernels[i][j]=clCreateKernel(programs[i][j],"rk2", &err);
            assert(err==CL_SUCCESS);
            
            free(full_program_source);
        }
    }
    
  
    
    size_t element_buffer_size;
    
    cl_mem cell_mem[num_element_arrays];
    cl_mem x_mem[num_element_arrays];
    cl_mem y_mem[num_element_arrays];
    cl_mem z_mem[num_element_arrays];
    
    cl_mem x_F1_mem[num_element_arrays];
    cl_mem y_F1_mem[num_element_arrays];
    cl_mem z_F1_mem[num_element_arrays];
    
    cl_mem params_mem;
    
   
    

#ifdef CELL_GROWTH
    //allocate memory on graphics card to allow for cell growth
    num_elements[num_element_arrays-1]=max_elements;

#endif


#ifdef CPU_CELL_GROWTH
    int temp_num_elements[num_element_arrays];
    num_elements[num_element_arrays-1]=0;
    for( int i = 0; i<num_element_arrays; i++)
    {
        temp_num_elements[i]=num_elements[i];
        num_elements[i]=max_elements;
    }
#endif
    

    for(int i = 0; i<num_element_arrays; i++)
    {
        
        
        element_buffer_size = num_elements[i]*sizeof(float);
        cell_mem[i]=clCreateBuffer(context, CL_MEM_READ_WRITE, num_elements[i]*sizeof(int), NULL, &err);
        assert(err==CL_SUCCESS);
        err = clEnqueueWriteBuffer(cmd_queue, cell_mem[i], CL_TRUE, 0, num_elements[i]*sizeof(int), (void*)cell_num[i], 0, NULL, NULL);
        assert(err==CL_SUCCESS);
        
        
        
        x_mem[i]=clCreateBuffer(context, CL_MEM_READ_WRITE, element_buffer_size, NULL, NULL);
        err = clEnqueueWriteBuffer(cmd_queue, x_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcellx[i], 0, NULL, NULL);
        assert(err==CL_SUCCESS);
         
        
        y_mem[i]=clCreateBuffer(context, CL_MEM_READ_WRITE, element_buffer_size, NULL, NULL);
        
        err = clEnqueueWriteBuffer(cmd_queue, y_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcelly[i], 0, NULL, NULL);
        
        assert(err==CL_SUCCESS);
       
        z_mem[i]=clCreateBuffer(context, CL_MEM_READ_WRITE, element_buffer_size, NULL, NULL);
        err = clEnqueueWriteBuffer(cmd_queue, z_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcellz[i], 0, NULL, NULL);
        assert(err==CL_SUCCESS);
        
        x_F1_mem[i]=clCreateBuffer(context, CL_MEM_READ_WRITE, element_buffer_size, NULL, NULL); 
        err = clEnqueueCopyBuffer(cmd_queue, x_mem[i], x_F1_mem[i], 0, 0, element_buffer_size, 0, NULL, NULL);
        assert(err==CL_SUCCESS);
        
        y_F1_mem[i]=clCreateBuffer(context, CL_MEM_READ_WRITE, element_buffer_size, NULL, NULL);
        err = clEnqueueCopyBuffer(cmd_queue, y_mem[i], y_F1_mem[i], 0, 0, element_buffer_size, 0, NULL, NULL);
        assert(err==CL_SUCCESS);
        
        z_F1_mem[i]=clCreateBuffer(context, CL_MEM_READ_WRITE, element_buffer_size, NULL, NULL);
        err = clEnqueueCopyBuffer(cmd_queue, z_mem[i], z_F1_mem[i], 0, 0, element_buffer_size, 0, NULL, NULL);
        assert(err==CL_SUCCESS);
        

    }
    
#ifdef CELL_GROWTH
    //after memory allocation reset number of elements to zero
    num_elements[num_element_arrays-1]=0;
    int num_stem_cells = 0;
    int new_elements;
    int growth_element_type = 2;
    int division_number = 0;
   
    for( int i = 0; i<num_elements[0]; i++)
    {
        if(cell_num[0][i]>num_stem_cells)num_stem_cells=cell_num[0][i];
    }
    num_stem_cells++;
    size_t global_cell_growth_work_size=num_stem_cells;
    cl_mem num_grown_elements_mem;
    cl_mem cell_division_true_mem;
    
    // the following 3 arrays are used later
    int num_grown_elements[num_stem_cells];
    int temp_grown_elements[num_stem_cells];
    int cell_division_true[num_stem_cells];
    
    for(int i = 0; i <num_stem_cells; i++)
    {
        num_grown_elements[i]=0;
        temp_grown_elements[i]=0;
    }
    num_grown_elements_mem=clCreateBuffer(context, CL_MEM_READ_WRITE, num_stem_cells*sizeof(int), NULL, NULL);
    cell_division_true_mem=clCreateBuffer(context, CL_MEM_READ_WRITE, num_stem_cells*sizeof(int), NULL, NULL);
    err = clEnqueueWriteBuffer(cmd_queue, num_grown_elements_mem, CL_TRUE, 0, num_stem_cells*sizeof(int), (void*)num_grown_elements, 0, NULL, NULL);
     
#endif
    
#ifdef CPU_CELL_GROWTH
    for( int i = 0; i<num_element_arrays; i++)
    {
        num_elements[i]=temp_num_elements[i];
    }
    //next_cell will be the cell number of the next cell that is added to the system
    int next_cell=0;
    int number_of_divisions[1024];
    for( int i = 0; i<1024; i++)
    {
        number_of_divisions[i]=0;
    }
    // This code for calculating basal cells only works if there are no super basal cells present
    // in the initial setup
    int num_basal_cells;
    // Set the number of supra basal cells to zero
    int num_supra_basal_cells = 0;
    for( int i = 0;i<num_element_arrays-1;i++)
    {
        for( int j = 0; j<num_elements[i];j++)
        {
            if( cell_num[i][j] > next_cell ) 
            {
                next_cell = cell_num[i][j];
            }
        }
    }
    next_cell++;
    num_basal_cells=next_cell;
    int next_supra_basal_cell=1000;
#endif
    
    

    params_mem=clCreateBuffer(context, CL_MEM_READ_ONLY, num_params*sizeof(float), NULL, NULL);
    err = clEnqueueWriteBuffer(cmd_queue, params_mem, CL_TRUE, 0, num_params*sizeof(float), (void*)params, 0, NULL, NULL);
    assert(err==CL_SUCCESS);
    

    
    
    size_t  local_work_size = DEFAULT_LOCAL_WORK_SIZE;
    size_t  global_work_size[num_element_arrays];
    size_t  actual_number_of_elements; // this will be used for kernels that do not need padding
    
    
    //pad the global work size so it is a multiple of the local work size
    
#ifndef RUN_ON_CPU
    size_t  shared_coordinate_sizes[num_element_arrays];//make sure you know how this works before modifying it
    size_t  shared_cell_num_sizes[num_element_arrays];
    
   

#endif
    

    int type_1;
    int type_2;
    char temp_str[32];
    char temp2[1024];

    int iterations_per_growth = ITERATIONS_BETWEEN_GROWTH_CYCLES;
    

    for( int itr=0; itr<max_itr; itr++)
    {

        if(itr%iterations_per_frame==0)
        {
            for(int i = 0; i<num_element_arrays; i++)
            {
                element_buffer_size = num_elements[i]*sizeof(float);
                
                err = clEnqueueReadBuffer(cmd_queue, cell_mem[i], CL_TRUE, 0, num_elements[i]*sizeof(int), (void*)cell_num[i], 0, NULL, NULL);
                assert(err==CL_SUCCESS);
                
                
                err = clEnqueueReadBuffer(cmd_queue, x_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcellx[i], 0, NULL, NULL);
                assert(err==CL_SUCCESS);
                
                
                err = clEnqueueReadBuffer(cmd_queue, y_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcelly[i], 0, NULL, NULL);
                assert(err==CL_SUCCESS);
                
                
                err = clEnqueueReadBuffer(cmd_queue, z_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcellz[i], 0, NULL, NULL);
                assert(err==CL_SUCCESS);
                
                
            }
            
            
            strcpy(temp2,output_name);
            sprintf(temp_str,"%d.pqr",itr);
            strcat(temp2,temp_str);
            FILE* fp;
            fp=fopen(temp2, "w");
            write_positions_pqr(fp, num_element_arrays, subcellx, subcelly, subcellz, cell_num, num_elements, element_type);
        }
      
#ifdef CELL_GROWTH
        if(itr%iterations_per_growth==0 && itr != 0)
        {
            
            int i = 0;
            int j = 2;
            err = clSetKernelArg(cell_growth_kernel, 0, sizeof(cl_mem), &cell_mem[i]);
            assert(err==CL_SUCCESS);
            err |= clSetKernelArg(cell_growth_kernel, 1, sizeof(cl_mem), &x_mem[i]);
            assert(err==CL_SUCCESS);
            err |= clSetKernelArg(cell_growth_kernel, 2, sizeof(cl_mem), &y_mem[i]);
            assert(err==CL_SUCCESS);
            err |= clSetKernelArg(cell_growth_kernel, 3, sizeof(cl_mem), &z_mem[i]);
            assert(err==CL_SUCCESS);
            err |= clSetKernelArg(cell_growth_kernel, 4, sizeof(int), &num_elements[i]);
            assert(err==CL_SUCCESS);
            err |= clSetKernelArg(cell_growth_kernel, 5, sizeof(cl_mem), &cell_mem[j]);
            assert(err==CL_SUCCESS);
            err |= clSetKernelArg(cell_growth_kernel, 6, sizeof(cl_mem), &x_mem[j]);
            assert(err==CL_SUCCESS);
            err |= clSetKernelArg(cell_growth_kernel, 7, sizeof(cl_mem), &y_mem[j]);
            assert(err==CL_SUCCESS);
            err |= clSetKernelArg(cell_growth_kernel, 8, sizeof(cl_mem), &z_mem[j]);
            assert(err==CL_SUCCESS);
            err |= clSetKernelArg(cell_growth_kernel, 9, sizeof(int), &num_elements[j]);
            assert(err==CL_SUCCESS);
            err |=clSetKernelArg(cell_growth_kernel, 10, sizeof(cl_mem), &num_grown_elements_mem),
            err = clEnqueueNDRangeKernel(cmd_queue, cell_growth_kernel, 1, NULL, &global_cell_growth_work_size, &local_work_size, 0, NULL, NULL);
            assert(err==CL_SUCCESS);
            err = clEnqueueReadBuffer(cmd_queue, num_grown_elements_mem, CL_TRUE, 0, num_stem_cells*sizeof(int), (void*)temp_grown_elements, 0, NULL, NULL);
            new_elements = 0;
            for(int k = 0; k < num_stem_cells; k++)
            {
                if(num_grown_elements[k] != temp_grown_elements[k])
                {
                    new_elements++;
                    num_grown_elements[k]=temp_grown_elements [k];
                }
            }
            num_elements[j]=num_elements[j]+new_elements;
            global_work_size[j]=num_elements[j];
            
            
            for(int k = 0; k<num_stem_cells; k++)
            {
                cell_division_true[k] = 0;
            }
            clEnqueueWriteBuffer(cmd_queue, cell_division_true_mem, CL_TRUE, 0, num_stem_cells*sizeof(int), (void*)cell_division_true, 0, NULL, NULL);
            
            time(&timer);
            random_seed = (unsigned int)timer;
        
            printf("Performing Cell Division\n");
            err = clSetKernelArg(cell_division_kernel, 0, sizeof(cl_mem), &cell_mem[growth_element_type]);
            printf("%d\n",err);
            err = clSetKernelArg(cell_division_kernel, 1, sizeof(int), &num_stem_cells);
            err = clSetKernelArg(cell_division_kernel, 2, sizeof(cl_mem), &num_grown_elements_mem);
            err = clSetKernelArg(cell_division_kernel, 3, sizeof(cl_mem), &cell_division_true_mem);
            err = clSetKernelArg(cell_division_kernel, 4, sizeof(int), &random_seed);
            err = clSetKernelArg(cell_division_kernel, 5, sizeof(int), &division_number);
            
            
            printf("%d\n",err);
            
            
            assert(err==CL_SUCCESS);
            err = clEnqueueNDRangeKernel(cmd_queue, cell_division_kernel, 1, NULL, &global_work_size[j], &local_work_size, 0, NULL, NULL);
            printf("%d\n",err);
            assert(err==CL_SUCCESS);
            
            clEnqueueReadBuffer(cmd_queue, cell_division_true_mem, CL_TRUE, 0, num_stem_cells*sizeof(int), (void*)cell_division_true, 0, NULL, NULL);
            
            for(int k = 0; k<num_stem_cells; k++)
            {
                if(cell_division_true[k]==1) 
                {
                    printf("%d\n",num_grown_elements[k]);
                    num_grown_elements[k]=0;
                    printf("a cell divided\n");
                    
                }      
            }
            err = clEnqueueWriteBuffer(cmd_queue, num_grown_elements_mem, CL_TRUE, 0, num_stem_cells*sizeof(int), (void*)num_grown_elements, 0, NULL, NULL);
            division_number++;
            
            
            
            
        }
#endif
          
            
#ifdef CPU_CELL_GROWTH
        if(itr%iterations_per_growth==0 && itr != 0)
        {
            //Copy element data from parrellel memory to cpu memory
            for(int i = 0; i<num_element_arrays; i++)
            {
                element_buffer_size = num_elements[i]*sizeof(float);
                
                err = clEnqueueReadBuffer(cmd_queue, cell_mem[i], CL_TRUE, 0, num_elements[i]*sizeof(int), (void*)cell_num[i], 0, NULL, NULL);
                assert(err==CL_SUCCESS);
                
                
                err = clEnqueueReadBuffer(cmd_queue, x_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcellx[i], 0, NULL, NULL);
                assert(err==CL_SUCCESS);
                
                
                err = clEnqueueReadBuffer(cmd_queue, y_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcelly[i], 0, NULL, NULL);
                assert(err==CL_SUCCESS);
                
                
                err = clEnqueueReadBuffer(cmd_queue, z_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcellz[i], 0, NULL, NULL);
                assert(err==CL_SUCCESS);
                
                //printf("Reading Array %d from graphics card\n",i);
                
                
            }
            
            //Cell division code goes here 
            
#ifdef SUPRA_BASAL_GROWTH
            //Calculate cell centers for supra basal cells
            int supra_basal_cell_id[num_supra_basal_cells];
            int supra_basal_cell_elements[num_supra_basal_cells];
            float supra_basal_cell_center_x[num_supra_basal_cells];
            float supra_basal_cell_center_y[num_supra_basal_cells];
            float supra_basal_cell_center_z[num_supra_basal_cells];
            float supra_basal_cell_max_z[num_supra_basal_cells];
            
            //the following is a measure of how many basal cells are close to this one
            //float supra_basal_cell_proximity[num_supra_basal_cells];
            
            for(int i = 0; i<num_supra_basal_cells; i++)
            {
                
                supra_basal_cell_elements[i]=0;
                supra_basal_cell_center_x[i]=0;
                supra_basal_cell_center_y[i]=0;
                supra_basal_cell_center_z[i]=0;
                supra_basal_cell_max_z[i]=0;
            }
            //get some information about the first cell to start things off
            //there might be a more elegant way of doing this
            //but it shouldn't effect the efficiency
        
            int found_cells = 0;
            int current_cell;
            int current_cell_id;
            //start this loop at 1 since we already took care of the first one
            for( int i = 0; i<num_elements[2]; i++)
            {
                //Set the current cell to be the next higher cell number
                current_cell=found_cells;
                current_cell_id = cell_num[2][i];
                if( current_cell_id < 1000 )
                {
                    continue;
                }
                for( int j = 0; j<found_cells; j++)
                {
                    //if the current cell already has data stored for it, point to that instead
                    printf("supra_basal_cell_id is %d and current cell id is %d\n",supra_basal_cell_id[j],current_cell_id);
                    if( supra_basal_cell_id[j]==current_cell_id )
                    {
                        //printf("setting current cell to %d\n",j);
                        current_cell=j;
                    }
                }
                if(current_cell==found_cells)
                {
                    found_cells++;
                }
                //This will do nothing if the cell number is already stored, but it
                // is probably more efficient then an if statement
                supra_basal_cell_id[current_cell]=cell_num[2][i];
                //printf("is this a valid supra_basal_cell_id? %d\n",supra_basal_cell_id[current_cell]);
                
                supra_basal_cell_elements[current_cell]++;
                supra_basal_cell_center_x[current_cell]+=subcellx[2][i];
                supra_basal_cell_center_y[current_cell]+=subcelly[2][i];
                supra_basal_cell_center_z[current_cell]+=subcellz[2][i];
                if(subcellz[2][i]>supra_basal_cell_max_z[current_cell])
                {
                    supra_basal_cell_max_z[current_cell]=subcellz[2][i];
                }
            }
            //This is a very innefficient way of doing this since cells need to be searched twice. There are much better ways but this is easy to code.
            for( int i= 0; i<num_supra_basal_cells; i++)
            {
                if(rand() % 1000 < 4)
                {
                    delete_cell(supra_basal_cell_id[i],num_elements,subcellx,subcelly,subcellz,cell_num);
                    printf("Deleting cell %d\n",supra_basal_cell_id[i]);
                    num_supra_basal_cells--;
                
                    i--;
                    continue;
                }
            }
            for(int i = 0; i<num_supra_basal_cells; i++)
            {
                
                supra_basal_cell_elements[i]=0;
                supra_basal_cell_center_x[i]=0;
                supra_basal_cell_center_y[i]=0;
                supra_basal_cell_center_z[i]=0;
                supra_basal_cell_max_z[i]=0;
            }
            //get some information about the first cell to start things off
            //there might be a more elegant way of doing this
            //but it shouldn't effect the efficiency
            
            found_cells = 0;
            //start this loop at 1 since we already took care of the first one
            for( int i = 0; i<num_elements[2]; i++)
            {
                //Set the current cell to be the next higher cell number
                current_cell=found_cells;
                current_cell_id = cell_num[2][i];
                if( current_cell_id < 1000 )
                {
                    continue;
                }
                for( int j = 0; j<found_cells; j++)
                {
                    //if the current cell already has data stored for it, point to that instead
                    printf("supra_basal_cell_id is %d and current cell id is %d\n",supra_basal_cell_id[j],current_cell_id);
                    if( supra_basal_cell_id[j]==current_cell_id )
                    {
                        //printf("setting current cell to %d\n",j);
                        current_cell=j;
                    }
                }
                if(current_cell==found_cells)
                {
                    found_cells++;
                }
                //This will do nothing if the cell number is already stored, but it
                // is probably more efficient then an if statement
                supra_basal_cell_id[current_cell]=cell_num[2][i];
                //printf("is this a valid supra_basal_cell_id? %d\n",supra_basal_cell_id[current_cell]);
                
                supra_basal_cell_elements[current_cell]++;
                supra_basal_cell_center_x[current_cell]+=subcellx[2][i];
                supra_basal_cell_center_y[current_cell]+=subcelly[2][i];
                supra_basal_cell_center_z[current_cell]+=subcellz[2][i];
                if(subcellz[2][i]>supra_basal_cell_max_z[current_cell])
                {
                    supra_basal_cell_max_z[current_cell]=subcellz[2][i];
                }
            }
            for( int i = 0; i<num_supra_basal_cells; i++)
            {
                supra_basal_cell_center_x[i]=supra_basal_cell_center_x[i]/supra_basal_cell_elements[i];
                supra_basal_cell_center_y[i]=supra_basal_cell_center_y[i]/supra_basal_cell_elements[i];
                supra_basal_cell_center_z[i]=supra_basal_cell_center_z[i]/supra_basal_cell_elements[i];
#ifdef CELL_CENTER_OUTPUT
                
                FILE *cell_center_data;
                cell_center_data = fopen("center_output.txt","a");
                fprintf(cell_center_data, "%f %f 1 %d\n", supra_basal_cell_center_x[i], supra_basal_cell_center_y[i], itr);
                fclose(cell_center_data);
#endif
                
            }
            if(found_cells != num_supra_basal_cells)
            {
                printf("found cells (%d) not equal to num_supra_basal_cells (%d), there is a problem.\n",found_cells,num_supra_basal_cells);
            }
      
#endif
            float temp_x,temp_y,temp_z;
#ifdef SUPRA_BASAL_GROWTH
            printf("num_supra_basal_cells is %d, iteration number is %d\n",num_supra_basal_cells,itr);
                      
            for( int i = 0; i<num_supra_basal_cells; i++)
            {
                
                
                
                temp_x=supra_basal_cell_center_x[i]*supra_basal_cell_elements[i];
                temp_y=supra_basal_cell_center_y[i]*supra_basal_cell_elements[i];
                temp_z=supra_basal_cell_center_z[i]*supra_basal_cell_elements[i];
                
                
                
                //printf("this is the current supra_basal_cell_id %d\n", supra_basal_cell_id[i]);
                printf("next cell num is: %d\n",supra_basal_cell_id[i]);
                cell_num[2][num_elements[2]]=supra_basal_cell_id[i];
                //printf("This is a test of random numbers %f %f\n",(float)rand()/RAND_MAX,(float)rand()/RAND_MAX);
                subcellx[2][num_elements[2]]=supra_basal_cell_center_x[i]+0.4*((float)rand()/RAND_MAX-0.5f);
                temp_x+=subcellx[2][num_elements[2]];
                
                subcelly[2][num_elements[2]]=supra_basal_cell_center_y[i]+0.4*((float)rand()/RAND_MAX-0.5f);
                temp_y+=subcelly[2][num_elements[2]];
                
                subcellz[2][num_elements[2]]=supra_basal_cell_center_z[i]+0.4*((float)rand()/RAND_MAX-0.5f);
                temp_z+=subcellz[2][num_elements[2]];
                
                num_elements[2]++;
                
                supra_basal_cell_elements[i]++;
                
                supra_basal_cell_center_x[i]=temp_x/supra_basal_cell_elements[i];
                supra_basal_cell_center_y[i]=temp_y/supra_basal_cell_elements[i];
                supra_basal_cell_center_z[i]=temp_z/supra_basal_cell_elements[i];
                
//the following is a hack and should be removed as soon as it is no longer needed
                
                //printf("this is the current supra_basal_cell_id %d\n", supra_basal_cell_id[i]);
                printf("next cell num is: %d\n",supra_basal_cell_id[i]);
                cell_num[2][num_elements[2]]=supra_basal_cell_id[i];
                //printf("This is a test of random numbers %f %f\n",(float)rand()/RAND_MAX,(float)rand()/RAND_MAX);
                subcellx[2][num_elements[2]]=supra_basal_cell_center_x[i]+0.4*((float)rand()/RAND_MAX-0.5f);
                temp_x+=subcellx[2][num_elements[2]];
                
                subcelly[2][num_elements[2]]=supra_basal_cell_center_y[i]+0.4*((float)rand()/RAND_MAX-0.5f);
                temp_y+=subcelly[2][num_elements[2]];
                
                subcellz[2][num_elements[2]]=supra_basal_cell_center_z[i]+0.4*((float)rand()/RAND_MAX-0.5f);
                temp_z+=subcellz[2][num_elements[2]];
                num_elements[2]++;
                
                supra_basal_cell_elements[i]++;
                
                supra_basal_cell_center_x[i]=temp_x/supra_basal_cell_elements[i];
                supra_basal_cell_center_y[i]=temp_y/supra_basal_cell_elements[i];
                supra_basal_cell_center_z[i]=temp_z/supra_basal_cell_elements[i];
                
                
                
                
            }
#endif
#ifdef SUPRA_BASAL_GROWTH
            int old_num_supra_basal_cells=num_supra_basal_cells;
            for(int i = 0; i<old_num_supra_basal_cells; i++)
            {
                
                if(supra_basal_cell_elements[i]*((float)rand()/RAND_MAX)>8.0f || supra_basal_cell_elements[i]>15)
                {
                    //divide along the x direction if rand()%2=0
                    int random_number = rand();
                    //printf("Is this number random? %d\n",random_number);
                    if(random_number%2==0)
                    {
                        //printf("dividing along x direction, num supra basal elements is %d\n", supra_basal_cell_elements[i]);
                        
                        int n=supra_basal_cell_elements[i];
                        int element_list[n];
                        int sorted_list[n];
                        int current_element;
                        int j = 2;
                        current_element = 0;
                        for(int k=0; k<num_elements[j];k++)
                        {
                            if(cell_num[j][k]==supra_basal_cell_id[i])
                            {
                                element_list[current_element]=k;
                                //printf("unsorted %d\n",element_list[current_element]);
                                current_element++;
                            }
                        }
                        
                        //now we sort this list;
                        sorted_list[0]=element_list[0];
                        for(int k=1; k<current_element; k++)
                        {
                            sorted_list[k]=element_list[k];
                            for(int itr = 0; itr<k; itr++)
                            {
                                if(subcellx[j][element_list[k]]<subcellx[j][sorted_list[itr]])
                                {
                                    for(int itr2 = k-1; itr2>=itr; itr2--)
                                    {
                                        sorted_list[itr2+1]=sorted_list[itr2];
                                    }
                                    sorted_list[itr]=element_list[k];
                                    //printf("Terminating loop %d\n", k);
                                    break;
                                }
                                //printf("Continuing loop %d\n", k);
                            }
                        }
                        //printf("Dividing along x %d\n", current_element);
                        for(int k=0; k<current_element/2; k++)
                        {
                            cell_num[j][sorted_list[k]]=next_supra_basal_cell;
                            //printf("sorted %d\n",sorted_list[k]);
                        }
                        
                        
                        
                    }
                    else //divide along the y direction
                    {
                        //printf("dividing along y direction, num supra basal elements is %d\n", supra_basal_cell_elements[i]);
                        int n=supra_basal_cell_elements[i];
                        int element_list[n];
                        int sorted_list[n];
                        int current_element;
                        int j=2;
                        current_element = 0;
                        for(int k=0; k<num_elements[j];k++)
                        {
                            if(cell_num[j][k]==supra_basal_cell_id[i])
                            {
                                element_list[current_element]=k;
                                //printf("unsorted %d\n",element_list[current_element]);
                                current_element++;
                            }
                        }
                        
                        //now we sort this list;
                        sorted_list[0]=element_list[0];
                        for(int k=1; k<current_element; k++)
                        {
                            sorted_list[k]=element_list[k];
                            for(int itr = 0; itr<k; itr++)
                            {
                                if(subcelly[j][element_list[k]]<subcelly[j][sorted_list[itr]])
                                {
                                    for(int itr2 = k-1; itr2>=itr; itr2--)
                                    {
                                        sorted_list[itr2+1]=sorted_list[itr2];
                                    }
                                    sorted_list[itr]=element_list[k];
                                    //printf("Terminating loop\n");
                                    break;
                                }
                                //printf("Continuing loop\n");
                            }
                        }
                        //printf("Dividing along y %d\n", current_element);
                        for(int k=0; k<current_element/2; k++)
                        {
                            cell_num[j][sorted_list[k]]=next_supra_basal_cell;
                            //printf("sorted %d\n",sorted_list[k]);
                        }
                        
                        
                    }
                    //printf("adding one to num_supra_basal_cells at supra basal division.\n");
                    next_supra_basal_cell++;
                                 
                    num_supra_basal_cells++;
                }
                
                
            }
            
#endif
            
            //Calculate cell centers for basal cells
            int basal_cell_id[num_basal_cells];
            int basal_cell_elements[num_basal_cells];
            float basal_cell_center_x[num_basal_cells];
            float basal_cell_center_y[num_basal_cells];
            float basal_cell_center_z[num_basal_cells];
            float basal_cell_max_z[num_basal_cells];
            
            //the following is a measure of how many basal cells are close to this one
            float basal_cell_proximity[num_basal_cells];
            for(int i = 0; i<num_basal_cells; i++)
            {
                basal_cell_elements[i]=0;
                basal_cell_center_x[i]=0;
                basal_cell_center_y[i]=0;
                basal_cell_center_z[i]=0;
                basal_cell_max_z[i]=0;
                basal_cell_proximity[i]=0;
            }
            //get some information about the first cell to start things off
            //there might be a more elegant way of doing this
            //but it shouldn't effect the efficiency
            basal_cell_id[0]=cell_num[0][0];
            basal_cell_elements[0]=1;
            basal_cell_center_x[0]=subcellx[0][0];
            basal_cell_center_y[0]=subcelly[0][0];
            basal_cell_center_z[0]=subcellz[0][0];
            found_cells = 1;
            //start this loop at 1 since we already took care of the first one
            for( int i = 1; i<num_elements[0]; i++)
            {
                //Set the current cell to be the next higher cell number
                current_cell=found_cells;
                for( int j = 0; j<found_cells; j++)
                {
                    //if the current cell already has data stored for it, point to that instead
                    if( basal_cell_id[j]==cell_num[0][i] )
                    {
                        current_cell=j;
                    }
                }
                if( current_cell==found_cells)
                {
                    found_cells++;
                }
                //This will do nothing if the cell number is already stored, but it 
                // is probably more efficient then an if statement
                basal_cell_id[current_cell]=cell_num[0][i];
                
                
                
                basal_cell_elements[current_cell]++;
                basal_cell_center_x[current_cell]+=subcellx[0][i];
                basal_cell_center_y[current_cell]+=subcelly[0][i];
                basal_cell_center_z[current_cell]+=subcellz[0][i];
                if(subcellz[0][i]>basal_cell_max_z[current_cell])
                {
                    basal_cell_max_z[current_cell]=subcellz[0][i];
                }
            }
            for( int i = 0; i<num_elements[0]; i++)
            {
                //Set the current cell to be the next higher cell number
                current_cell=found_cells;
                for( int j = 0; j<found_cells; j++)
                {
                    //if the current cell already has data stored for it, point to that instead
                    if( basal_cell_id[j]==cell_num[1][i] )
                    {
                        current_cell=j;
                    }
                }
                if( current_cell==found_cells)
                {
                    found_cells++;
                }
                //This will do nothing if the cell number is already stored, but it 
                // is more efficient then an if statement
                basal_cell_id[current_cell]=cell_num[1][i];
                
                
                basal_cell_elements[current_cell]++;
                basal_cell_center_x[current_cell]+=subcellx[1][i];
                basal_cell_center_y[current_cell]+=subcelly[1][i];
                basal_cell_center_z[current_cell]+=subcellz[1][i];
                if(subcellz[1][i]>basal_cell_max_z[current_cell])
                {
                    basal_cell_max_z[current_cell]=subcellz[1][i];
                }
            }
            for(int i = 0; i<num_basal_cells; i++)
            {
                if(rand() % 1000 < 4)
                {
                    delete_cell(basal_cell_id[i],num_elements,subcellx,subcelly,subcellz,cell_num);
                    printf("Deleting cell %d\n",basal_cell_id[i]);
                    num_basal_cells--;
                   
                    i--;
                    continue;
                }
            }
            for(int i = 0; i<num_basal_cells; i++)
            {
                basal_cell_elements[i]=0;
                basal_cell_center_x[i]=0;
                basal_cell_center_y[i]=0;
                basal_cell_center_z[i]=0;
                basal_cell_max_z[i]=0;
                basal_cell_proximity[i]=0;
            }
            //get some information about the first cell to start things off
            //there might be a more elegant way of doing this
            //but it shouldn't effect the efficiency
            basal_cell_id[0]=cell_num[0][0];
            basal_cell_elements[0]=1;
            basal_cell_center_x[0]=subcellx[0][0];
            basal_cell_center_y[0]=subcelly[0][0];
            basal_cell_center_z[0]=subcellz[0][0];
            found_cells = 1;
            //start this loop at 1 since we already took care of the first one
            for( int i = 1; i<num_elements[0]; i++)
            {
                //Set the current cell to be the next higher cell number
                current_cell=found_cells;
                for( int j = 0; j<found_cells; j++)
                {
                    //if the current cell already has data stored for it, point to that instead
                    if( basal_cell_id[j]==cell_num[0][i] )
                    {
                        current_cell=j;
                    }
                }
                if( current_cell==found_cells)
                {
                    found_cells++;
                }
                //This will do nothing if the cell number is already stored, but it
                // is probably more efficient then an if statement
                basal_cell_id[current_cell]=cell_num[0][i];
                
                
                
                basal_cell_elements[current_cell]++;
                basal_cell_center_x[current_cell]+=subcellx[0][i];
                basal_cell_center_y[current_cell]+=subcelly[0][i];
                basal_cell_center_z[current_cell]+=subcellz[0][i];
                if(subcellz[0][i]>basal_cell_max_z[current_cell])
                {
                    basal_cell_max_z[current_cell]=subcellz[0][i];
                }
            }
            for( int i = 0; i<num_elements[0]; i++)
            {
                //Set the current cell to be the next higher cell number
                current_cell=found_cells;
                for( int j = 0; j<found_cells; j++)
                {
                    //if the current cell already has data stored for it, point to that instead
                    if( basal_cell_id[j]==cell_num[1][i] )
                    {
                        current_cell=j;
                    }
                }
                if( current_cell==found_cells)
                {
                    found_cells++;
                }
                //This will do nothing if the cell number is already stored, but it
                // is more efficient then an if statement
                basal_cell_id[current_cell]=cell_num[1][i];
                
                
                basal_cell_elements[current_cell]++;
                basal_cell_center_x[current_cell]+=subcellx[1][i];
                basal_cell_center_y[current_cell]+=subcelly[1][i];
                basal_cell_center_z[current_cell]+=subcellz[1][i];
                if(subcellz[1][i]>basal_cell_max_z[current_cell])
                {
                    basal_cell_max_z[current_cell]=subcellz[1][i];
                }
            }

            for( int i = 0; i<num_basal_cells; i++)
            {
                basal_cell_center_x[i]=basal_cell_center_x[i]/basal_cell_elements[i];
                basal_cell_center_y[i]=basal_cell_center_y[i]/basal_cell_elements[i];
                basal_cell_center_z[i]=basal_cell_center_z[i]/basal_cell_elements[i];
#ifdef DELAMINATION
                    
                if(basal_cell_center_z[i]>DELAMINATION_RANGE)
                {
                    
#ifdef TIME_TO_DETACHMENT
                    FILE *detachment_data;
                    detachment_data = fopen("ttfd.txt","a");
                    fprintf(detachment_data,"%d\n", itr);
                    fclose(detachment_data);
                    
                    return;
#endif
#ifdef CELL_CENTER_OUTPUT
                    //the basal cell has become a supra basal cell, but the cell center remains the same, so we output it as a supra basal cell
                
                    FILE *cell_center_data;
                    cell_center_data = fopen("center_output.txt","a");
                    fprintf(cell_center_data, "%f %f 1 %d\n", basal_cell_center_x[i], basal_cell_center_y[i], itr);
                    fclose(cell_center_data);
#endif
                    cell_delaminate(basal_cell_id[i], num_element_types, num_element_arrays, element_type, max_elements, num_elements, subcellx, subcelly, subcellz, cell_num,next_supra_basal_cell);
                    for(int j=i; j<num_basal_cells-1; j++)
                    {
                        basal_cell_id[j]=basal_cell_id[j+1];
                        basal_cell_elements[j]=basal_cell_elements[j+1];
                        basal_cell_center_x[j]=basal_cell_center_x[j+1];
                        basal_cell_center_y[j]=basal_cell_center_y[j+1];
                        basal_cell_center_z[j]=basal_cell_center_z[j+1];
                            
                    }
                    num_basal_cells--;
                    //printf("adding one to num_supra_basal_cells at delamination.\n");
                    next_supra_basal_cell++;
                                  
                    num_supra_basal_cells++;


                    i--;
                    continue;
                }

#endif
#ifdef CELL_CENTER_OUTPUT
                
                FILE *cell_center_data;
                cell_center_data = fopen("center_output.txt","a");
                fprintf(cell_center_data, "%f %f 0 %d\n", basal_cell_center_x[i], basal_cell_center_y[i], itr);
                fclose(cell_center_data);
#endif
 

            }
            
#ifndef GLOBAL_GROWTH_SWITCH
            float d;
            
            for( int i = 0; i<num_basal_cells; i++)
            {
                for( int j = 0; j<i; j++)
                {
                    d=distance(basal_cell_center_x[i],basal_cell_center_y[i],basal_cell_center_z[i],basal_cell_center_x[j],basal_cell_center_y[j],basal_cell_center_z[j]);
                    basal_cell_proximity[i]+=1.0f/(d*d);
                }
                for( int j = i+1; j<num_basal_cells; j++)
                {
                    d=distance(basal_cell_center_x[i],basal_cell_center_y[i],basal_cell_center_z[i],basal_cell_center_x[j],basal_cell_center_y[j],basal_cell_center_z[j]);
                    basal_cell_proximity[i]+=1.0f/(d*d);
                }
            }
            
            
            
            for( int i = 0; i<num_basal_cells; i++)
            {
                printf("Cell %d proximity number is %f\n", basal_cell_id[i], basal_cell_proximity[i]);
            }
#else
            //printf("Number of basal cells and elements: %d, %d, %d\n",num_basal_cells,num_elements[0],num_elements[1]);
            
#endif  

            for( int i = 0; i<num_basal_cells; i++)
            {
#ifndef GLOBAL_GROWTH_SWITCH
                if(basal_cell_proximity[i]<0.39f)
                {
#else
                    
                if(num_basal_cells<MAX_BASAL_CELLS)
                {
#endif
                    temp_x=basal_cell_center_x[i]*basal_cell_elements[i];
                    temp_y=basal_cell_center_y[i]*basal_cell_elements[i];
                    temp_z=basal_cell_center_z[i]*basal_cell_elements[i];
                        
                    cell_num[0][num_elements[0]]=basal_cell_id[i];
                    //printf("This is a test of random numbers %f %f\n",(float)rand()/RAND_MAX,(float)rand()/RAND_MAX);
                    subcellx[0][num_elements[0]]=basal_cell_center_x[i]+0.5*((float)rand()/RAND_MAX-0.5f);
                    temp_x+=subcellx[0][num_elements[0]];
                        
                    subcelly[0][num_elements[0]]=basal_cell_center_y[i]+0.5*((float)rand()/RAND_MAX-0.5f);
                    temp_y+=subcelly[0][num_elements[0]];
                        
                    subcellz[0][num_elements[0]]=basal_cell_center_z[i]+2.0f+0.5*((float)rand()/RAND_MAX-0.5f);
                    temp_z+=subcellz[0][num_elements[0]];
                        
                    num_elements[0]++;
                                            
                        
                    cell_num[1][num_elements[1]]=basal_cell_id[i];
                        
                    subcellx[1][num_elements[1]]=basal_cell_center_x[i]+0.5*((float)rand()/RAND_MAX-0.5f);
                    temp_x+=subcellx[1][num_elements[1]];
                        
                    subcelly[1][num_elements[1]]=basal_cell_center_y[i]+0.5*((float)rand()/RAND_MAX-0.5f);
                    temp_y+=subcelly[1][num_elements[1]];
                        
                    subcellz[1][num_elements[1]]=basal_cell_center_z[i]-2.0f+0.5*((float)rand()/RAND_MAX-0.5f);
                        temp_z+=subcellz[1][num_elements[1]];
                        
                    num_elements[1]++;
                    basal_cell_elements[i]+=2;
                        
                    basal_cell_center_x[i]=temp_x/basal_cell_elements[i];
                    basal_cell_center_y[i]=temp_y/basal_cell_elements[i];
                    basal_cell_center_z[i]=temp_z/basal_cell_elements[i];
                        
                }
                else
                {
                    if(num_elements[2]<(max_elements-1))
                    {
                        float temp_x;
                        float temp_y;
                        float temp_z;
                        temp_x=basal_cell_center_x[i]*basal_cell_elements[i];
                        temp_y=basal_cell_center_y[i]*basal_cell_elements[i];
                        temp_z=basal_cell_center_z[i]*basal_cell_elements[i];
                       
                        printf("cell num is: %d\n",basal_cell_id[i]);
                        cell_num[2][num_elements[2]]=basal_cell_id[i];
                        //printf("This is a test of random numbers %f %f\n",(float)rand()/RAND_MAX,(float)rand()/RAND_MAX);
                        subcellx[2][num_elements[2]]=basal_cell_center_x[i]+0.5f*((float)rand()/RAND_MAX-0.5f);
                        //temp_x+=subcellx[2][num_elements[2]];
                       
                        subcelly[2][num_elements[2]]=basal_cell_center_y[i]+0.5f*((float)rand()/RAND_MAX-0.5f);
                        //temp_y+=subcelly[2][num_elements[2]];
                       
                        subcellz[2][num_elements[2]]=basal_cell_max_z[i]+1.0f+0.25f*((float)rand()/RAND_MAX-0.5f);
                        //temp_z+=subcellz[2][num_elements[2]];
                       
                        num_elements[2]++;
                       
                        
                        //basal_cell_elements[i]+=1;
                        
//                        basal_cell_center_x[i]=temp_x/basal_cell_elements[i];
//                        basal_cell_center_y[i]=temp_y/basal_cell_elements[i];
//                        basal_cell_center_z[i]=temp_z/basal_cell_elements[i];
//                        
                        // This is a hack, remove as soon as it is no longer necessary
                        printf("cell num is: %d\n",basal_cell_id[i]);
                        cell_num[2][num_elements[2]]=basal_cell_id[i];
                        //printf("This is a test of random numbers %f %f\n",(float)rand()/RAND_MAX,(float)rand()/RAND_MAX);
                        subcellx[2][num_elements[2]]=basal_cell_center_x[i]+0.5f*((float)rand()/RAND_MAX-0.5f);
                        //temp_x+=subcellx[2][num_elements[2]];
                        
                        subcelly[2][num_elements[2]]=basal_cell_center_y[i]+0.5f*((float)rand()/RAND_MAX-0.5f);
                        //temp_y+=subcelly[2][num_elements[2]];
                        
                        subcellz[2][num_elements[2]]=basal_cell_max_z[i]+1.0f+0.25f*((float)rand()/RAND_MAX-0.5f);
                        //temp_z+=subcellz[2][num_elements[2]];
                        
                        num_elements[2]++;
                        
                        
                        //basal_cell_elements[i]+=1;
                        
//                        basal_cell_center_x[i]=temp_x/basal_cell_elements[i];
//                        basal_cell_center_y[i]=temp_y/basal_cell_elements[i];
//                        basal_cell_center_z[i]=temp_z/basal_cell_elements[i];
                
                    }

                }
            }
            //Divide basal cells if the are large enough
                
            int current_num_basal_cells=num_basal_cells;
                
            for(int i=0; i<current_num_basal_cells; i++)
            {
                
#ifndef GLOBAL_GROWTH_SWITCH
                if((basal_cell_elements[i]*(1.0f*rand())/RAND_MAX > 8.0f) || basal_cell_elements[i] > 15)
                {
#else
                if((num_basal_cells<MAX_BASAL_CELLS && basal_cell_elements[i]*(1.0f*rand())/RAND_MAX > 8.0f) || basal_cell_elements[i] > 15)
                {
#endif   
                    //divide along the x direction if rand()%2=0
                    int random_number = rand();
                    //printf("Is this number random? %d\n",random_number);
                    if(random_number%2==0)
                    {
                                    
                        int n=basal_cell_elements[i];
                        int element_list[n];
                        int sorted_list[n];
                        int current_element;
                        for(int j=0; j<2;j++)
                        {
                            current_element = 0;
                            for(int k=0; k<num_elements[j];k++)
                            {
                                if(cell_num[j][k]==basal_cell_id[i])
                                {
                                    element_list[current_element]=k;
                                    //printf("unsorted %d\n",element_list[current_element]);
                                    current_element++;
                                }
                            }
                                    
                            //now we sort this list;
                            sorted_list[0]=element_list[0];
                            for(int k=1; k<current_element; k++)
                            {
                                sorted_list[k]=element_list[k];
                                for(int itr = 0; itr<k; itr++)
                                {
                                    if(subcellx[j][element_list[k]]<subcellx[j][sorted_list[itr]])
                                    {
                                        for(int itr2 = k-1; itr2>=itr; itr2--)
                                        {
                                            sorted_list[itr2+1]=sorted_list[itr2];
                                        }
                                        sorted_list[itr]=element_list[k];
                                        //printf("Terminating loop %d\n", k);
                                        break;
                                    }
                                    //printf("Continuing loop %d\n", k);
                                }
                            }
                            //printf("Dividing along x %d\n", current_element);
                            for(int k=0; k<current_element/2; k++)
                            {
                                cell_num[j][sorted_list[k]]=next_cell;
                                //printf("sorted %d\n",sorted_list[k]);
                            }
                                    
                                    
                        }
                    }
                    else //divide along the y direction
                    {
                        int n=basal_cell_elements[i];
                        int element_list[n];
                        int sorted_list[n];
                        int current_element;
                        for(int j=0; j<2;j++)
                        {
                            current_element = 0;
                            for(int k=0; k<num_elements[j];k++)
                            {
                                if(cell_num[j][k]==basal_cell_id[i])
                                {
                                    element_list[current_element]=k;
                                    //printf("unsorted %d\n",element_list[current_element]);
                                    current_element++;
                                }
                            }
                                    
                                    //now we sort this list;
                            sorted_list[0]=element_list[0];
                            for(int k=1; k<current_element; k++)
                            {
                                sorted_list[k]=element_list[k];
                                for(int itr = 0; itr<k; itr++)
                                {
                                    if(subcelly[j][element_list[k]]<subcelly[j][sorted_list[itr]])
                                    {
                                        for(int itr2 = k-1; itr2>=itr; itr2--)
                                        {
                                            sorted_list[itr2+1]=sorted_list[itr2];
                                        }
                                        sorted_list[itr]=element_list[k];
                                        //printf("Terminating loop\n");
                                        break;
                                    }
                                    //printf("Continuing loop\n");
                                }
                            }
                            //printf("Dividing along y %d\n", current_element);
                            for(int k=0; k<current_element/2; k++)
                            {
                                cell_num[j][sorted_list[k]]=next_cell;
                                //printf("sorted %d\n",sorted_list[k]);
                            }
                                    
                                    
                        }
                                
                    }
                    next_cell++;
                    num_basal_cells++;
                            
                            
                }
            }
                    
                    
            int num_type2_cells[num_basal_cells];
            for(int i = 0; i<num_basal_cells; i++)
            {
                num_type2_cells[i]=0;
            }
                    
            for(int i = 0; i<num_elements[2]; i++)
            {
                for(int j = 0; j<num_basal_cells; j++)
                {
                    if(basal_cell_id[j]==cell_num[2][i])
                    {
                        num_type2_cells[j]++;
                    }
                }
            }
                    
            for(int i = 0; i<num_basal_cells; i++)
            {
                if(num_type2_cells[i]*((float)rand())/RAND_MAX>8 || num_type2_cells[i]>15)
                {
                    for(int j = 0; j<num_elements[2]; j++)
                    {
                        if(cell_num[2][j]==basal_cell_id[i])
                        {
                            printf("cell num is: %d\n",next_supra_basal_cell);
                            cell_num[2][j]=next_supra_basal_cell;
                        }
                    }
                    //printf("adding one to num_supra_basal_cells at asymmetric division.\n");
                    next_supra_basal_cell++;
                    
                    num_supra_basal_cells++;
                    number_of_divisions[i]++;
                }
            }
                
                
                    
            //Copy element data from CPU memory to parrallel memory
            for(int i = 0; i<num_element_arrays; i++)
            {
                element_buffer_size = num_elements[i]*sizeof(float);
                        
                err = clEnqueueWriteBuffer(cmd_queue, cell_mem[i], CL_TRUE, 0, num_elements[i]*sizeof(int), (void*)cell_num[i], 0, NULL, NULL);
                assert(err==CL_SUCCESS);
                        
                        
                err = clEnqueueWriteBuffer(cmd_queue, x_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcellx[i], 0, NULL, NULL);
                assert(err==CL_SUCCESS);
                        
                        
                err = clEnqueueWriteBuffer(cmd_queue, y_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcelly[i], 0, NULL, NULL);
                assert(err==CL_SUCCESS);
                      
                        
                err = clEnqueueWriteBuffer(cmd_queue, z_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcellz[i], 0, NULL, NULL);
                assert(err==CL_SUCCESS);
                        
                //printf("Writing array %d to graphics card\n",i);
                
                        
            }
                    
#ifdef CELL_NUMBER_OUTPUT
                FILE *output_data;
                output_data = fopen("cell_output.txt","a");
                fprintf(output_data,"%d %d %d\n", num_basal_cells, num_supra_basal_cells, itr);
                fclose(output_data);
#endif
        
                
    }

        
            
            
#endif
                
                
#ifndef RUN_ON_CPU
        for(int i = 0; i<num_element_arrays; i++)
        {
//            if(num_elements[i]==0)
//            {
//                shared_coordinate_sizes[i]=3*sizeof(float);
//                shared_cell_num_sizes[i]=sizeof(int);
//                        
//            }
            {
                shared_coordinate_sizes[i]=3*num_elements[i]*sizeof(float);
                shared_cell_num_sizes[i]=num_elements[i]*sizeof(int);
            }
            if(num_elements[i]%local_work_size == 0)
            {
                global_work_size[i]=num_elements[i];
            }
            else
            {
                global_work_size[i]=((num_elements[i]/local_work_size)+1)*local_work_size;
            }
                
        }
        for(int i = 0; i<num_element_arrays; i++)
        {
            for(int j = 0; j<num_element_arrays; j++)
            {
                if((num_elements[i]>0)&&(num_elements[j]>0))
                {
                    //find out element types to decide which kernel to use;
                    type_1 = element_type[i];
                    type_2 = element_type[j];
                    //printf("number of type %d elements is %d\n",type_1,num_elements[i]);
                    
                    err = clSetKernelArg(kernels[type_1][type_2], 0, sizeof(cl_mem), &cell_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 1, sizeof(cl_mem), &x_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 2, sizeof(cl_mem), &y_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 3, sizeof(cl_mem), &z_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 4, sizeof(int), &num_elements[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 5, sizeof(cl_mem), &cell_mem[j]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 6, sizeof(cl_mem), &x_mem[j]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 7, sizeof(cl_mem), &y_mem[j]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 8, sizeof(cl_mem), &z_mem[j]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 9, sizeof(int), &num_elements[j]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 10, sizeof(float), &dt);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 11, shared_coordinate_sizes[j], NULL);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 12, shared_cell_num_sizes[j], NULL);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 13, sizeof(cl_mem), &x_F1_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 14, sizeof(cl_mem), &y_F1_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 15, sizeof(cl_mem), &z_F1_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 16, sizeof(cl_mem), &params_mem);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 17, sizeof(int), &num_params);
                    assert(err==CL_SUCCESS);
                    
                    //size_t thread_size;
                    //clGetKernelWorkGroupInfo(kernels[type_1][type_2], devices, CL_KERNEL_WORK_GROUP_SIZE, sizeof(size_t), &thread_size, NULL);
                    //printf("Recommended Size: %lu\n", thread_size);
                    
                    err = clEnqueueNDRangeKernel(cmd_queue, kernels[type_1][type_2], 1, NULL, &global_work_size[i], &local_work_size, 0, NULL, NULL);
                    //printf("The error code is %d, type 1 is %d, type 2 is %d, global work size is %lu, and local work size is %d\n",err, type_1, type_2, global_work_size[i],local_work_size);
                    assert(err==CL_SUCCESS);
                }
                
            }
            
        }
        // this may not be a good place to insert random noise, it is however the fastest place
        
        
        
        /*
         for(int i = 0; i<num_element_arrays; i++)
         {
         element_buffer_size = num_elements[i]*sizeof(float);
         err = clEnqueueCopyBuffer(cmd_queue, x_F1_mem[i], x_mem[i], 0, 0, element_buffer_size, 0, 0, NULL);
         assert(err == CL_SUCCESS);
         err = clEnqueueCopyBuffer(cmd_queue, y_F1_mem[i], y_mem[i], 0, 0, element_buffer_size, 0, 0, NULL);
         assert(err == CL_SUCCESS);
         err = clEnqueueCopyBuffer(cmd_queue, z_F1_mem[i], z_mem[i], 0, 0, element_buffer_size, 0, 0, NULL);
         assert(err == CL_SUCCESS);
         }
         */
        
        
        for(int i = 0; i<num_element_arrays; i++)
        {
            for(int j = 0; j<num_element_arrays; j++)
            {
                if((num_elements[i]>0)&&(num_elements[j]>0))
                {
                    //find out element types to decide which kernel to use;
                    type_1 = element_type[i];
                    type_2 = element_type[j];
                    
                    err = clSetKernelArg(kernels[type_1][type_2], 0, sizeof(cl_mem), &cell_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 1, sizeof(cl_mem), &x_F1_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 2, sizeof(cl_mem), &y_F1_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 3, sizeof(cl_mem), &z_F1_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 4, sizeof(int), &num_elements[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 5, sizeof(cl_mem), &cell_mem[j]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 6, sizeof(cl_mem), &x_F1_mem[j]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 7, sizeof(cl_mem), &y_F1_mem[j]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 8, sizeof(cl_mem), &z_F1_mem[j]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 9, sizeof(int), &num_elements[j]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 10, sizeof(float), &dt_2);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 11, shared_coordinate_sizes[j], NULL);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 12, shared_cell_num_sizes[j], NULL);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 13, sizeof(cl_mem), &x_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 14, sizeof(cl_mem), &y_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 15, sizeof(cl_mem), &z_mem[i]);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 16, sizeof(cl_mem), &params_mem);
                    assert(err==CL_SUCCESS);
                    err |= clSetKernelArg(kernels[type_1][type_2], 17, sizeof(int), &num_params);
                    assert(err==CL_SUCCESS);
                    
                    //size_t thread_size;
                    //clGetKernelWorkGroupInfo(kernels[type_1][type_2], devices, CL_KERNEL_WORK_GROUP_SIZE, sizeof(size_t), &thread_size, NULL);
                    //printf("Recommended Size: %lu\n", thread_size);
                    
                    err = clEnqueueNDRangeKernel(cmd_queue, kernels[type_1][type_2], 1, NULL, &global_work_size[i], &local_work_size, 0, NULL, NULL);
                    assert(err==CL_SUCCESS);
                }
                
            }
            
        }
        
        for(int i = 0; i<num_element_arrays; i++)
        {   
            if(num_elements[i]>0)
            {
                actual_number_of_elements = (size_t)num_elements[i];
                time(&timer);
                random_seed = (unsigned int)timer;
                err = clSetKernelArg(noisy_copy_kernel, 0, sizeof(cl_mem), &x_F1_mem[i]);    
                assert(err == CL_SUCCESS);
                err = clSetKernelArg(noisy_copy_kernel, 1, sizeof(cl_mem), &y_F1_mem[i]);    
                assert(err == CL_SUCCESS);
                err = clSetKernelArg(noisy_copy_kernel, 2, sizeof(cl_mem), &z_F1_mem[i]);    
                assert(err == CL_SUCCESS);
                err = clSetKernelArg(noisy_copy_kernel, 3, sizeof(cl_mem), &x_F1_mem[i]);    
                assert(err == CL_SUCCESS);
                err = clSetKernelArg(noisy_copy_kernel, 4, sizeof(cl_mem), &y_F1_mem[i]);    
                assert(err == CL_SUCCESS);
                err = clSetKernelArg(noisy_copy_kernel, 5, sizeof(cl_mem), &z_F1_mem[i]);    
                assert(err == CL_SUCCESS);
                err = clSetKernelArg(noisy_copy_kernel, 6, sizeof(unsigned int), &random_seed);    
                assert(err == CL_SUCCESS);
                err = clSetKernelArg(noisy_copy_kernel, 7, sizeof(float), &lower_bound);    
                assert(err == CL_SUCCESS);
                err = clSetKernelArg(noisy_copy_kernel, 8, sizeof(float), &upper_bound);    
                assert(err == CL_SUCCESS);
                
                err = clEnqueueNDRangeKernel(cmd_queue, noisy_copy_kernel, 1, NULL, &actual_number_of_elements, &local_work_size, 0, NULL, NULL);
                
                
                
                
                
                
                element_buffer_size = num_elements[i]*sizeof(float);
                err = clEnqueueCopyBuffer(cmd_queue, x_mem[i], x_F1_mem[i], 0, 0, element_buffer_size, 0, 0, NULL);
                assert(err == CL_SUCCESS);
                err = clEnqueueCopyBuffer(cmd_queue, y_mem[i], y_F1_mem[i], 0, 0, element_buffer_size, 0, 0, NULL);
                assert(err == CL_SUCCESS);
                err = clEnqueueCopyBuffer(cmd_queue, z_mem[i], z_F1_mem[i], 0, 0, element_buffer_size, 0, 0, NULL);
                assert(err == CL_SUCCESS);
            }
        }
    
                
#else
        for(int i = 0; i<num_element_arrays; i++)
        {
            global_work_size[i]=num_elements[i];
        }        
        for(int i = 0; i<num_element_arrays; i++)
        {
            
            for(int j = 0; j<num_element_arrays; j++)
            {
                //find out element types to decide which kernel to use;
                type_1 = element_type[i];
                type_2 = element_type[j];
                err = clSetKernelArg(kernels[type_1][type_2], 0, sizeof(cl_mem), &cell_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 1, sizeof(cl_mem), &x_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 2, sizeof(cl_mem), &y_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 3, sizeof(cl_mem), &z_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 4, sizeof(int), &num_elements[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 5, sizeof(cl_mem), &cell_mem[j]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 6, sizeof(cl_mem), &x_mem[j]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 7, sizeof(cl_mem), &y_mem[j]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 8, sizeof(cl_mem), &z_mem[j]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 9, sizeof(int), &num_elements[j]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 10, sizeof(float), &dt);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 11, sizeof(cl_mem), &x_F1_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 12, sizeof(cl_mem), &y_F1_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 13, sizeof(cl_mem), &z_F1_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 14, sizeof(cl_mem), &params_mem);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 15, sizeof(int), &num_params);
                assert(err==CL_SUCCESS);
                
                //size_t thread_size;
                //clGetKernelWorkGroupInfo(kernels[type_1][type_2], devices, CL_KERNEL_WORK_GROUP_SIZE, sizeof(size_t), &thread_size, NULL);
                //printf("Recommended Size: %lu\n", thread_size);
                if(num_elements[i]!=0)
                {
                    err = clEnqueueNDRangeKernel(cmd_queue, kernels[type_1][type_2], 1, NULL, &global_work_size[i], &local_work_size, 0, NULL, NULL);
                    assert(err==CL_SUCCESS);
                }
                
            }
            
            
        }
        /*
         for(int i = 0; i<num_element_arrays; i++)
         {
         element_buffer_size = num_elements[i]*sizeof(float);
         err = clEnqueueCopyBuffer(cmd_queue, x_F1_mem[i], x_mem[i], 0, 0, element_buffer_size, 0, 0, NULL);
         assert(err == CL_SUCCESS);
         err = clEnqueueCopyBuffer(cmd_queue, y_F1_mem[i], y_mem[i], 0, 0, element_buffer_size, 0, 0, NULL);
         assert(err == CL_SUCCESS);
         err = clEnqueueCopyBuffer(cmd_queue, z_F1_mem[i], z_mem[i], 0, 0, element_buffer_size, 0, 0, NULL);
         assert(err == CL_SUCCESS);
         }
         */
        
        
        
        for(int i = 0; i<num_element_arrays; i++)
        {
            for(int j = 0; j<num_element_arrays; j++)
            {
                //find out element types to decide which kernel to use;
                type_1 = element_type[i];
                type_2 = element_type[j];
                
                err = clSetKernelArg(kernels[type_1][type_2], 0, sizeof(cl_mem), &cell_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 1, sizeof(cl_mem), &x_F1_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 2, sizeof(cl_mem), &y_F1_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 3, sizeof(cl_mem), &z_F1_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 4, sizeof(int), &num_elements[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 5, sizeof(cl_mem), &cell_mem[j]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 6, sizeof(cl_mem), &x_F1_mem[j]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 7, sizeof(cl_mem), &y_F1_mem[j]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 8, sizeof(cl_mem), &z_F1_mem[j]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 9, sizeof(int), &num_elements[j]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 10, sizeof(float), &dt_2);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 11, sizeof(cl_mem), &x_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 12, sizeof(cl_mem), &y_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 13, sizeof(cl_mem), &z_mem[i]);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 14, sizeof(cl_mem), &params_mem);
                assert(err==CL_SUCCESS);
                err |= clSetKernelArg(kernels[type_1][type_2], 15, sizeof(int), &num_params);
                assert(err==CL_SUCCESS);
                
                //size_t thread_size;
                //clGetKernelWorkGroupInfo(kernels[type_1][type_2], devices, CL_KERNEL_WORK_GROUP_SIZE, sizeof(size_t), &thread_size, NULL);
                //printf("Recommended Size: %lu\n", thread_size);
                if( num_elements[i]!=0)
                {
                    err = clEnqueueNDRangeKernel(cmd_queue, kernels[type_1][type_2], 1, NULL, &global_work_size[i], &local_work_size, 0, NULL, NULL);
                    assert(err==CL_SUCCESS);
                }
                
            }
            
        }
        
        
        
        for(int i = 0; i<num_element_arrays; i++)
        {
            
            actual_number_of_elements = (size_t)num_elements[i];
            time(&timer);
            random_seed = (unsigned int)timer;
            err = clSetKernelArg(noisy_copy_kernel, 0, sizeof(cl_mem), &x_mem[i]);    
            assert(err == CL_SUCCESS);
            err = clSetKernelArg(noisy_copy_kernel, 1, sizeof(cl_mem), &y_mem[i]);    
            assert(err == CL_SUCCESS);
            err = clSetKernelArg(noisy_copy_kernel, 2, sizeof(cl_mem), &z_mem[i]);    
            assert(err == CL_SUCCESS);
            err = clSetKernelArg(noisy_copy_kernel, 3, sizeof(cl_mem), &x_mem[i]);    
            assert(err == CL_SUCCESS);
            err = clSetKernelArg(noisy_copy_kernel, 4, sizeof(cl_mem), &y_mem[i]);    
            assert(err == CL_SUCCESS);
            err = clSetKernelArg(noisy_copy_kernel, 5, sizeof(cl_mem), &z_mem[i]);    
            assert(err == CL_SUCCESS);
            err = clSetKernelArg(noisy_copy_kernel, 6, sizeof(unsigned int), &random_seed);    
            assert(err == CL_SUCCESS);
            err = clSetKernelArg(noisy_copy_kernel, 7, sizeof(float), &lower_bound);    
            assert(err == CL_SUCCESS);
            err = clSetKernelArg(noisy_copy_kernel, 8, sizeof(float), &upper_bound);    
            assert(err == CL_SUCCESS);
            
            err = clEnqueueNDRangeKernel(cmd_queue, noisy_copy_kernel, 1, NULL, &actual_number_of_elements, &local_work_size, 0, NULL, NULL);
            
            
            
            element_buffer_size = num_elements[i]*sizeof(float);
            err = clEnqueueCopyBuffer(cmd_queue, x_mem[i], x_F1_mem[i], 0, 0, element_buffer_size, 0, 0, NULL);
            assert(err == CL_SUCCESS);
            err = clEnqueueCopyBuffer(cmd_queue, y_mem[i], y_F1_mem[i], 0, 0, element_buffer_size, 0, 0, NULL);
            assert(err == CL_SUCCESS);
            err = clEnqueueCopyBuffer(cmd_queue, z_mem[i], z_F1_mem[i], 0, 0, element_buffer_size, 0, 0, NULL);
            assert(err == CL_SUCCESS);
            
        }
        
#endif
    }
    
    
    
    
    
    for(int i = 0; i<num_element_arrays; i++)
    {
        element_buffer_size = num_elements[i]*sizeof(float);
        
        err = clEnqueueReadBuffer(cmd_queue, cell_mem[i], CL_TRUE, 0, num_elements[i]*sizeof(int), (void*)cell_num[i], 0, NULL, NULL);
        assert(err==CL_SUCCESS);
        
        
        err = clEnqueueReadBuffer(cmd_queue, x_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcellx[i], 0, NULL, NULL);
        assert(err==CL_SUCCESS);
        
       
        err = clEnqueueReadBuffer(cmd_queue, y_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcelly[i], 0, NULL, NULL);
        assert(err==CL_SUCCESS);
        
    
        err = clEnqueueReadBuffer(cmd_queue, z_mem[i], CL_TRUE, 0, element_buffer_size, (void*)subcellz[i], 0, NULL, NULL);
        assert(err==CL_SUCCESS);
        //printf("Copied arrays for cell %d to CPU memory.\n",i);    
        
    }

    for(int i = 0; i<num_element_types; i++)
    {
        for(int j = 0; j<num_element_types; j++)
        {
            clReleaseKernel(kernels[i][j]);
            clReleaseProgram(programs[i][j]);
        }
    }
    
    clReleaseCommandQueue(cmd_queue);
    clReleaseContext(context);
    
    for(int i = 0; i<num_element_arrays; i++)
    {
        clReleaseMemObject(cell_mem[i]);
        clReleaseMemObject(x_mem[i]);
        clReleaseMemObject(y_mem[i]);
        clReleaseMemObject(z_mem[i]);
        clReleaseMemObject(x_F1_mem[i]);
        clReleaseMemObject(y_F1_mem[i]);
        clReleaseMemObject(z_F1_mem[i]);
    }
    
    clReleaseMemObject(params_mem);
    
    for(int i = 0; i<num_element_types; i++)
    {
        free(kernels[i]);
        free(programs[i]);
    }
    free(kernels);
    free(programs);

    
    
}

int main (int argc, const char * argv[])
{
   
    int num_element_types = DEFAULT_ELEMENT_TYPES;
    int num_element_arrays = num_element_types; //this may change later
    int max_elements = DEFAULT_MAX_ELEMENTS;
    int iterations_per_frame = 0; // by default do not record intermediate iterations
    char output_format[256];
    
    float dt = (float)DEFAULT_TIME_STEP;
    int itr = DEFAULT_ITERATIONS;
    int num_params=DEFAULT_NUM_PARAMS;
    char ***filenames;
    
    //allocate memory to hold cell information
    float **subcellx, **subcelly, **subcellz;
    int **cell_num;
    
    //create a pointer to the number of elements in each array
    int *num_elements;
    //create a pointer to the element type of each array
    int *element_type;
    
    float *params;

       
    printf("\n");
    if(argc < 3)
    {
        printf("No options file given, using default values\n");
    }
    else
    {
        printf("Using Options file: %s\n", argv[2]);
        read_options(argv[2], &itr, &max_elements, output_format, &iterations_per_frame);
    }

    printf("\n");
    
    if( argc == 1 )
    {
        printf("No inputs given, using default values\n");
        //create arrays of pointers to point to arrays of floats.
        subcellx=(float **)malloc(num_element_arrays*sizeof(float*));
        subcelly=(float **)malloc(num_element_arrays*sizeof(float*));
        subcellz=(float **)malloc(num_element_arrays*sizeof(float*));
        cell_num=(int **)malloc(num_element_arrays*sizeof(int*));
        num_elements=(int *)malloc(num_element_arrays*sizeof(int));
        element_type=(int *)malloc(num_element_arrays*sizeof(int));
        
        
        for(int i = 0; i<num_element_arrays; i++)
        {
            subcellx[i]=(float *)malloc(max_elements*sizeof(float));
            subcelly[i]=(float *)malloc(max_elements*sizeof(float));
            subcellz[i]=(float *)malloc(max_elements*sizeof(float));
            cell_num[i]=(int *)malloc(max_elements*sizeof(int));
        }
        
        //generate some garbage initial data for testing
        for( int i=0; i<num_element_arrays; i++)
        {
            element_type[i]=i;
            num_elements[i]=max_elements;
            for( int j=0; j<max_elements; j++)
            {
                subcellx[i][j]=j*1.0;;
                subcelly[i][j]=0.0;
                subcellz[i][j]=i*1.0;
                cell_num[i][j]=1;
            }
        }

    
    }
    else
    {
        printf("Using initial conditions in file: %s\n", argv[1]);
        read_initial_conditions(argv[1], &num_element_types, &num_element_arrays, max_elements, &subcellx, &subcelly, &subcellz, &cell_num, &num_elements, &element_type); 
        printf("%d element types detected\n", num_element_types);
        
    }
        
    filenames=(char ***)malloc(num_element_types*sizeof(char**));
    
    printf("\n");
     
    if(argc < 4 )
    {
        printf("No potentials file given, using default values\n");
        params = (float*)malloc(num_params*sizeof(float));
        //load the default parameters, this will be user modified later
        params[0]=(float)DEFAULT_U_INTRA;
        params[1]=(float)DEFAULT_UE_INTRA;
        params[2]=(float)DEFAULT_V_INTRA;
        params[3]=(float)DEFAULT_VE_INTRA;
        params[4]=(float)DEFAULT_U_INTRA;
        params[5]=(float)DEFAULT_UE_INTRA;
        params[6]=(float)DEFAULT_V_INTRA;
        params[7]=(float)DEFAULT_VE_INTRA;
  
         
           
        for(int i = 0; i<num_element_types; i++)
        {
            filenames[i]=(char **)malloc(num_element_types*sizeof(char*));

            
            for(int j = 0; j<num_element_types ; j++)
            {
                filenames[i][j]=DEFAULT_POTENTIAL;
            }
        }



    }
    else
    {
        
        for(int i = 0; i<num_element_types; i++)
        {
            filenames[i]=(char **)malloc(num_element_types*sizeof(char*));
            for(int j = 0; j<num_element_types ; j++)
            {
                filenames[i][j]=(char*)malloc(1024);
                strcpy(filenames[i][j],DEFAULT_NO_INTERACTION_PROGRAM);
            }
        }

        printf("Using potential functions given by file: %s\n", argv[3]);
        read_potential_functions(argv[3], &num_params, &params, num_element_types, filenames);
        
      
    }
     
    
    // the following code prints out the starting points of the elements
    
    /*
    for( int i=0; i<num_element_arrays; i++)
    {
        for( int j=0; j<num_elements[i]; j++)
        {
            printf("x=%2.2f, y=%2.2f, z=%2.2f, cell number=%i\n", subcellx[i][j], subcelly[i][j], subcellz[i][j], cell_num[i][j]);
        }
    }
    */
   
    
    exec_kernels(num_element_types, num_element_arrays, element_type, max_elements, num_elements, subcellx, subcelly, subcellz, cell_num, dt, itr, num_params, params, filenames, argv[4],iterations_per_frame);
 
    // the following code prints out the ending points of the elements
    
    
    if(argc < 5)
    {
        printf("No outputfile given, printing to stdout\n");
        for( int i=0; i<num_element_arrays; i++)
        {
            for( int j=0; j<num_elements[i]; j++)
            {
                printf("x=%2.2f, y=%2.2f, z=%2.2f, cell number=%i\n", subcellx[i][j], subcelly[i][j], subcellz[i][j], cell_num[i][j]);
            }
        }
    }
    else
    {
        printf("Writing output to: %s\n", argv[4]);
        FILE * fp;
        fp = fopen(argv[4], "w");
        if(strcmp(output_format,"GNUPLOT")==0)
        {
            printf("test3\n");
            write_positions_gnuplot(fp, num_element_arrays, subcellx, subcelly, subcellz, cell_num, num_elements, element_type);
        }
        else if(strcmp(output_format,"PQR")==0)
        {
            printf("test1\n");
            write_positions_pqr(fp, num_element_arrays, subcellx, subcelly, subcellz, cell_num, num_elements, element_type);
        }
        else
        {
        
            write_positions(fp, num_element_arrays, subcellx, subcelly, subcellz, cell_num, num_elements, element_type);
        }
        fclose(fp);
    }
    
    
    
    
    for( int i=0; i<num_element_arrays ; i++)
    {
        free(subcellx[i]);
        free(subcelly[i]);
        free(subcellz[i]);
        free(cell_num[i]);
    }
    
    for( int i = 0; i<num_element_types; i++)
    {
        free(filenames[i]);
    }
    
    free(subcellx);
    free(subcelly);
    free(subcellz);
    free(cell_num);
    free(filenames);
    
    return 0;
}

